

period_schema = {
    'month': ['year', 'month'],
    'week': ['year', 'month', 'week', ],
    'day': ['date'],
    'quarter': ['year', 'quarter']
}

resample_schema = {
    'second': 'S',
    'minute': 'T',
    'hour': 'H',
    'week': 'W',
    'day': 'D',
    'month': 'M',
    'quarter': 'Q'
}
